@@ -84,7 +84,7 @@ GRP <- function(X, ...) UseMethod("GRP") # , X
 # Added... could also do in GRP.default... but this is better, no match.call etc... match.call takes 4 microseconds. could do both ?? think about possible applications...
 GRP.GRP <- function(X, ...) X
 
-GRP.default <- function(X, by = NULL, sort = TRUE, decreasing = FALSE, na.last = TRUE,
+GRP.default <- function(X, by = NULL, sort = .op[["sort"]], decreasing = FALSE, na.last = TRUE,
                         return.groups = TRUE, return.order = sort, method = "auto",
                         call = TRUE, ...) {
 
@@ -192,8 +192,8 @@ GRPN <- function(x, expand = TRUE, ...) {
   if(expand) .Call(C_subsetVector, g$group.sizes, g$group.id, FALSE) else g$group.sizes
 }
 
-# Only for masking if collapse_mask = "all"
-n <- function(x, g, TRA, ...) {
+# dplyr-style n(): only for masking if collapse_mask = "all"
+n_internal <- function(x, g, TRA, ...) {
   if(missing(g)) {
     if(missing(x)) stop("if data is not grouped need to call n() on a column")
     return(if(is.list(x)) fnrow(x) else length(x))
@@ -253,7 +253,7 @@ plot.GRP <- function(x, breaks = "auto", type = "s", horizontal = FALSE, ...) {
   hist(x[[3L]], breaks, xlab = "Group Size", main = "Histogram of Group Sizes", ...)
 }
 
-as_factor_GRP <- function(x, ordered = FALSE) { # , ...
+as_factor_GRP <- function(x, ordered = FALSE, sep = ".") { # , ...
   # if(is.factor(x)) return(x)
   # if(!is_GRP(x)) stop("x must be a 'GRP' object")
   f <- x[[2L]]
@@ -264,7 +264,7 @@ as_factor_GRP <- function(x, ordered = FALSE) { # , ...
     if(length(gr) == 1L) {
       attr(f, "levels") <- tochar(gr[[1L]]) # or formatC ?
     } else {
-      attr(f, "levels") <- do.call(paste, c(gr, list(sep = ".")))
+      attr(f, "levels") <- do.call(paste, c(gr, list(sep = sep)))
     }
   }
   oldClass(f) <- if(ordered) c("ordered","factor","na.included") else c("factor","na.included") # previously if any(x[[6L]])
@@ -276,9 +276,9 @@ as.factor_GRP <- function(x, ordered = FALSE) {
   as_factor_GRP(x, ordered)
 }
 
-finteraction <- function(..., factor = TRUE, ordered = FALSE, sort = factor, method = "auto") { # does it drop levels ? -> Yes !
+finteraction <- function(..., factor = TRUE, ordered = FALSE, sort = factor && .op[["sort"]], method = "auto", sep = ".") { # does it drop levels ? -> Yes !
   X <- if(...length() == 1L && is.list(..1)) ..1 else list(...)
-  if(factor) return(as_factor_GRP(GRP.default(X, sort = sort, return.order = FALSE, method = method, call = FALSE), ordered))
+  if(factor) return(as_factor_GRP(GRP.default(X, sort = sort, return.order = FALSE, method = method, call = FALSE), ordered, sep))
   if(sort || method == "radix") {
     g <- GRP.default(X, sort = sort, return.groups = FALSE, return.order = FALSE, method = method, call = FALSE)
     res <- g[[2L]]
@@ -365,7 +365,7 @@ GRP.pseries <- function(X, effect = 1L, ..., group.sizes = TRUE, return.groups =
 GRP.pdata.frame <- function(X, effect = 1L, ..., group.sizes = TRUE, return.groups = TRUE, call = TRUE)
   GRP.pseries(X, effect, ..., group.sizes = group.sizes, return.groups = return.groups, call = call)
 
-fgroup_by <- function(.X, ..., sort = TRUE, decreasing = FALSE, na.last = TRUE, return.groups = TRUE, return.order = sort, method = "auto") {          #   e <- substitute(list(...)) # faster but does not preserve attributes of unique groups !
+fgroup_by <- function(.X, ..., sort = .op[["sort"]], decreasing = FALSE, na.last = TRUE, return.groups = TRUE, return.order = sort, method = "auto") {          #   e <- substitute(list(...)) # faster but does not preserve attributes of unique groups !
   clx <- oldClass(.X)
   oldClass(.X) <- NULL
   m <- match(c("GRP_df", "grouped_df", "data.frame"), clx, nomatch = 0L)
@@ -514,7 +514,7 @@ condCopyAttrib <- function(x, d) {
   if(is.object(x)) return(x)
   cld <- oldClass(d)
   condalcSA(x, list(names = attr(x, "names"),
-                    row.names = .set_row_names(length(x[[1L]])),
+                    row.names = .set_row_names(.Call(C_fnrow, x)),
                     class = cld[cld %!in% c("GRP_df", "grouped_df", "sf", "pdata.frame", "indexed_frame")]),
             any(cld == "data.table"))
   # attr(d, "groups") <- NULL
@@ -675,7 +675,7 @@ as.factor_qG <- function(x, ordered = FALSE, na.exclude = TRUE) {
   as_factor_qG(x, ordered, na.exclude)
 }
 
-qF <- function(x, ordered = FALSE, na.exclude = TRUE, sort = TRUE, drop = FALSE,
+qF <- function(x, ordered = FALSE, na.exclude = TRUE, sort = .op[["sort"]], drop = FALSE,
                keep.attr = TRUE, method = "auto") {
   if(is.factor(x) && sort) {
     if(!keep.attr && !all(names(ax <- attributes(x)) %in% c("levels", "class")))
@@ -702,7 +702,7 @@ qF <- function(x, ordered = FALSE, na.exclude = TRUE, sort = TRUE, drop = FALSE,
          stop("Unknown method:", method))
 }
 
-qG <- function(x, ordered = FALSE, na.exclude = TRUE, sort = TRUE,
+qG <- function(x, ordered = FALSE, na.exclude = TRUE, sort = .op[["sort"]],
                return.groups = FALSE, method = "auto") {
   if(inherits(x, c("factor", "qG"))) {
     nainc <- inherits(x, "na.included")
@@ -748,7 +748,7 @@ qG <- function(x, ordered = FALSE, na.exclude = TRUE, sort = TRUE,
 
 radixuniquevec <- function(x, sort, na.last = TRUE, decreasing = FALSE) {
   o <- .Call(C_radixsort, na.last, decreasing, TRUE, FALSE, sort, pairlist(x))
-  if(attr(o, "maxgrpn") == 1L && (!sort || attr(o, "sorted"))) return(x)
+  if(attr(o, "maxgrpn") <= 1L && (!sort || attr(o, "sorted"))) return(x)
   Csv(x, if(attr(o, "sorted")) attr(o, "starts") else Csv(o, attr(o, "starts")))
 }
 
@@ -771,7 +771,7 @@ funique.data.frame <- function(x, cols = NULL, sort = FALSE, method = "auto", ..
   # if(!missing(...)) unused_arg_action(match.call(), ...)
   use.group <- switch(method, auto = !sort, hash = TRUE, radix = FALSE, stop("method needs to be 'auto', 'hash' or 'radix'."))
   o <- switchGRP(if(is.null(cols)) x else colsubset(x, cols), starts = TRUE, sort = sort, use.group = use.group, ...)
-  if((use.group && length(o) == attr(o, "N.groups")) || (!use.group && attr(o, "maxgrpn") == 1L && (!sort || attr(o, "sorted")))) # return(x)
+  if((use.group && length(o) == attr(o, "N.groups")) || (!use.group && attr(o, "maxgrpn") <= 1L && (!sort || attr(o, "sorted")))) # return(x)
      return(if(inherits(x, "data.table")) alc(x) else x)
   st <- if(use.group || attr(o, "sorted")) attr(o, "starts") else Csv(o, attr(o, "starts"))
   rn <- attr(x, "row.names")
@@ -803,7 +803,7 @@ funique.sf <- function(x, cols = NULL, sort = FALSE, method = "auto", ...) {
   cols <- if(is.null(cols)) whichv(attr(x, "names"), attr(x, "sf_column"), TRUE) else
                             cols2int(cols, x, attr(x, "names"), FALSE)
   o <- switchGRP(.subset(x, cols), starts = TRUE, sort = sort, use.group = use.group, ...)
-  if((use.group && length(o) == attr(o, "N.groups")) || (!use.group && attr(o, "maxgrpn") == 1L && (!sort || attr(o, "sorted")))) return(x)
+  if((use.group && length(o) == attr(o, "N.groups")) || (!use.group && attr(o, "maxgrpn") <= 1L && (!sort || attr(o, "sorted")))) return(x)
   st <- if(use.group || attr(o, "sorted")) attr(o, "starts") else Csv(o, attr(o, "starts"))
   rn <- attr(x, "row.names")
   res <- .Call(C_subsetDT, x, st, seq_along(unclass(x)), FALSE)
@@ -816,7 +816,7 @@ funique.pseries <- function(x, sort = FALSE, method = "auto", drop.index.levels
   if(is.array(x)) stop("funique currently only supports atomic vectors and data.frames")
   use.group <- switch(method, auto = !sort, hash = TRUE, radix = FALSE, stop("method needs to be 'auto', 'hash' or 'radix'."))
   o <- switchGRP(x, starts = TRUE, sort = sort, use.group = use.group, ...)
-  if((use.group && length(o) == attr(o, "N.groups")) || (!use.group && attr(o, "maxgrpn") == 1L && (!sort || attr(o, "sorted")))) return(x)
+  if((use.group && length(o) == attr(o, "N.groups")) || (!use.group && attr(o, "maxgrpn") <= 1L && (!sort || attr(o, "sorted")))) return(x)
   st <- if(use.group || attr(o, "sorted")) attr(o, "starts") else Csv(o, attr(o, "starts"))
   res <- Csv(x, st)
   if(length(names(x))) names(res) <- Csv(names(x), st)
@@ -829,7 +829,7 @@ funique.pseries <- function(x, sort = FALSE, method = "auto", drop.index.levels
 funique.pdata.frame <- function(x, cols = NULL, sort = FALSE, method = "auto", drop.index.levels = "id", ...) {
   use.group <- switch(method, auto = !sort, hash = TRUE, radix = FALSE, stop("method needs to be 'auto', 'hash' or 'radix'."))
   o <- switchGRP(if(is.null(cols)) x else colsubset(x, cols), starts = TRUE, sort = sort, use.group = use.group, ...)
-  if((use.group && length(o) == attr(o, "N.groups")) || (!use.group && attr(o, "maxgrpn") == 1L && (!sort || attr(o, "sorted")))) # return(x)
+  if((use.group && length(o) == attr(o, "N.groups")) || (!use.group && attr(o, "maxgrpn") <= 1L && (!sort || attr(o, "sorted")))) # return(x)
     return(if(inherits(x, "data.table")) alc(x) else x)
   st <- if(use.group || attr(o, "sorted")) attr(o, "starts") else Csv(o, attr(o, "starts"))
   rn <- attr(x, "row.names")
@@ -843,10 +843,27 @@ funique.pdata.frame <- function(x, cols = NULL, sort = FALSE, method = "auto", d
 }
 
 fnunique <- function(x) {
+  if(is.list(x) && length(unclass(x)) == 1L) x <- .subset2(x, 1L)
   if(is.atomic(x) && !is.complex(x)) .Call(C_fndistinct, x, NULL, FALSE, 1L) else
     attr(.Call(C_group, x, FALSE, FALSE), "N.groups")
 }
 
+any_duplicated <- function(x) fnunique(x) < (if(is.atomic(x)) length(x) else .Call(C_fnrow, x))
+
+fduplicated <- function(x, all = FALSE) {
+  if(all) {
+    g <- .Call(C_group, x, FALSE, FALSE)
+    ng <- attr(g, "N.groups")
+    if(ng == length(g)) return(.Call(C_alloc, FALSE, length(g), TRUE))
+    gs <- .Call(C_fwtabulate, g, NULL, ng, FALSE)
+    return(.Call(C_subsetVector, gs != 1L, g, FALSE))
+  }
+  g <- .Call(C_group, x, TRUE, FALSE)
+  starts <- attr(g, "starts")
+  if(length(starts) == length(g)) return(.Call(C_alloc, FALSE, length(g), TRUE))
+  .Call(C_setcopyv, .Call(C_alloc, TRUE, length(g), TRUE), starts, FALSE, FALSE, TRUE, TRUE)
+}
+
 fdroplevels <- function(x, ...) UseMethod("fdroplevels")
 
 fdroplevels.default <- function(x, ...) {
