@@ -1,6 +1,3 @@
-#ifdef _OPENMP
-#include <omp.h>
-#endif
 #include "collapse_c.h"
 
 SEXP Cna_rm(SEXP x) {
@@ -46,7 +43,7 @@ SEXP Cna_rm(SEXP x) {
     return out;
   }
   case VECSXP: {
-    const SEXP *xd = SEXPPTR(x);
+    const SEXP *xd = SEXPPTR_RO(x);
     for (int i = 0; i != n; ++i) if(length(xd[i]) == 0) ++k;
     if(k == 0) return x;
     SEXP out = PROTECT(allocVector(VECSXP, n - k));
@@ -61,7 +58,7 @@ SEXP Cna_rm(SEXP x) {
   error("Unsupported type '%s' passed to na_rm()", type2char(TYPEOF(x)));
 }
 
-// Helper function to find a single sting in factor levels
+// Helper function to find a single string in factor levels
 int fchmatch(SEXP x, SEXP val, int nomatch) {
   const SEXP *px = STRING_PTR(x), v = asChar(val);
   for(int i = 0, l = length(x); i != l; ++i) if(px[i] == v) return i + 1;
@@ -97,28 +94,32 @@ if(length(val) == n && n > 1) {
     const int *px = INTEGER(x);
     const int *pv = INTEGER(val);
     WHICHVLOOPLX
-      break;
+    break;
   }
   case REALSXP:
   {
     const double *px = REAL(x);
     const double *pv = REAL(val);
-    WHICHVLOOPLX
-      break;
+    if(invert) {
+      for(int i = 0; i != n; ++i) if(px[i] != pv[i] && (NISNAN(px[i]) || NISNAN(pv[i]))) buf[j++] = i+1;
+    } else {
+      for(int i = 0; i != n; ++i) if(px[i] == pv[i] || (ISNAN(px[i]) && ISNAN(pv[i]))) buf[j++] = i+1;
+    }
+    break;
   }
   case STRSXP:
   {
     const SEXP *px = STRING_PTR(x);
     const SEXP *pv = STRING_PTR(val);
     WHICHVLOOPLX
-      break;
+    break;
   }
   case RAWSXP :
   {
     const Rbyte *px = RAW(x);
     const Rbyte *pv = RAW(val);
     WHICHVLOOPLX
-      break;
+    break;
   }
   default: error("Unsupported type '%s' passed to whichv()", type2char(TYPEOF(x)));
   }
@@ -135,7 +136,7 @@ if(length(val) == n && n > 1) {
       v = fchmatch(getAttrib(x, R_LevelsSymbol), val, 0);
     } else v = asInteger(val);
     WHICHVLOOP
-      break;
+    break;
   }
   case REALSXP:
   {
@@ -155,16 +156,17 @@ if(length(val) == n && n > 1) {
   case STRSXP:
   {
     const SEXP *px = STRING_PTR(x);
-    const SEXP v = asChar(val);
+    const SEXP v = PROTECT(asChar(val));
     WHICHVLOOP
-      break;
+    UNPROTECT(1);
+    break;
   }
   case RAWSXP :
   {
     const Rbyte *px = RAW(x);
     const Rbyte v = RAW(val)[0];
     WHICHVLOOP
-      break;
+    break;
   }
   default: error("Unsupported type '%s' passed to whichv()", type2char(TYPEOF(x)));
   }
@@ -179,6 +181,7 @@ return(ans);
 SEXP anyallv(SEXP x, SEXP val, SEXP Rall) {
 
   int n = length(x), all = asLogical(Rall);
+  if(length(x) == 0) return ScalarLogical(all ? 1 : 0);
   if(length(val) != 1) error("value needs to be length 1");
 
 #define ALLANYVLOOP                                                      \
@@ -201,7 +204,7 @@ case LGLSXP:
     v = fchmatch(getAttrib(x, R_LevelsSymbol), val, 0);
   } else v = asInteger(val);
   ALLANYVLOOP
-    break;
+  break;
 }
 case REALSXP:
 {
@@ -209,21 +212,21 @@ case REALSXP:
   const double v = asReal(val);
   if(ISNAN(v)) error("please use allNA()");
   ALLANYVLOOP
-    break;
+  break;
 }
 case STRSXP:
 {
   const SEXP *px = STRING_PTR(x);
   const SEXP v = asChar(val);
   ALLANYVLOOP
-    break;
+  break;
 }
 case RAWSXP :
 {
   const Rbyte *px = RAW(x);
   const Rbyte v = RAW(val)[0];
   ALLANYVLOOP
-    break;
+  break;
 }
 default: error("Unsupported type '%s' passed to allv() / anyv()", type2char(TYPEOF(x)));
 }
@@ -261,13 +264,15 @@ SEXP setcopyv(SEXP x, SEXP val, SEXP rep, SEXP Rinvert, SEXP Rset, SEXP Rind1) {
     if(lr != 1 && lr != n) error("If length(v) == 1, length(r) must be 1 or length(x)");
   }
 
-  if(lr > 1 && tr != tx) { // lr == n &&
+  if(tr != tx) { // lr == n &&
     if(!((tx == INTSXP && tr == LGLSXP) || (tx == LGLSXP && tr == INTSXP))) {
-      // PROTECT_INDEX ipx;
-      // PROTECT_WITH_INDEX(rep = coerceVector(rep, tx), &ipx);
-      tr = tx;
-      rep = PROTECT(coerceVector(rep, tx));
-      ++nprotect;
+      if(tr > tx && !(lr == 1 && tx == INTSXP && tr == REALSXP && REAL_ELT(rep, 0) == (int)REAL_ELT(rep, 0)))
+         warning("Type of R (%s) is larger than X (%s) and thus coerced. This incurs loss of information, such as digits of real numbers being truncated upon coercion to integer. To avoid this, make sure X has a larger type than R: character > double > integer > logical.", type2char(tr), type2char(tx));
+      if(lr > 1) {
+        tr = tx;
+        rep = PROTECT(coerceVector(rep, tx));
+        ++nprotect;
+      }
     } // error("typeof(x) needs to match typeof(r)");
   }
 
@@ -287,7 +292,8 @@ SEXP setcopyv(SEXP x, SEXP val, SEXP rep, SEXP Rinvert, SEXP Rset, SEXP Rind1) {
 
   #define setcopyvLOOPLVEC1                                     \
   if(tv == INTSXP) {                                            \
-    for(int i = 0; i != lv; ++i) px[pv[i]-1] = r;               \
+    _Pragma("omp simd")                                         \
+    for(int i = 0; i < lv; ++i) px[pv[i]-1] = r;                \
   } else if(invert == 0) {                                      \
     for(int i = 0; i != n; ++i) if(pv[i] > 0) px[i] = r;        \
   } else {                                                      \
@@ -297,9 +303,11 @@ SEXP setcopyv(SEXP x, SEXP val, SEXP rep, SEXP Rinvert, SEXP Rset, SEXP Rind1) {
   #define setcopyvLOOPLVEC                                      \
   if(tv == INTSXP) {                                            \
     if(lr == n) {                                               \
-      for(int i = 0; i != lv; ++i) px[pv[i]-1] = pr[pv[i]-1];   \
+      _Pragma("omp simd")                                       \
+      for(int i = 0; i < lv; ++i) px[pv[i]-1] = pr[pv[i]-1];    \
     } else {                                                    \
-      for(int i = 0; i != lv; ++i) px[pv[i]-1] = pr[i];         \
+      _Pragma("omp simd")                                       \
+      for(int i = 0; i < lv; ++i) px[pv[i]-1] = pr[i];          \
     }                                                           \
   } else if(invert == 0) {                                      \
     for(int i = 0; i != n; ++i) if(pv[i] > 0) px[i] = pr[i];    \
@@ -383,8 +391,9 @@ SEXP setcopyv(SEXP x, SEXP val, SEXP rep, SEXP Rinvert, SEXP Rset, SEXP Rind1) {
     if(lv == 1 && ind1 == 0) {
       const SEXP v = PROTECT(asChar(val));
       if(lr == 1) {
-        const SEXP r = asChar(rep);
+        const SEXP r = PROTECT(asChar(rep));
         setcopyvLOOP(r)
+        UNPROTECT(1);
       } else {
         const SEXP *restrict pr = STRING_PTR(rep);
         setcopyvLOOP(pr[i])
@@ -393,8 +402,9 @@ SEXP setcopyv(SEXP x, SEXP val, SEXP rep, SEXP Rinvert, SEXP Rset, SEXP Rind1) {
     } else {
       const int *restrict pv = INTEGER(val); // ALTREP(val) ? (const int *)ALTVEC_DATAPTR(val) :
       if(lr == 1) {
-        const SEXP r = asChar(rep);
+        const SEXP r = PROTECT(asChar(rep));
         setcopyvLOOPLVEC1
+        UNPROTECT(1);
       } else {
         const SEXP *restrict pr = STRING_PTR(rep);
         setcopyvLOOPLVEC
@@ -404,6 +414,7 @@ SEXP setcopyv(SEXP x, SEXP val, SEXP rep, SEXP Rinvert, SEXP Rset, SEXP Rind1) {
   }
   case VECSXP:
   {
+    if(set && ALTREP(x)) error("cannot modify ALTREP list by reference");
     SEXP *restrict px = set ? SEXPPTR(x) : SEXPPTR(ans);
     if(lv == 1 && ind1 == 0) error("Cannot compare lists to a value");
     // if(tr != VECSXP) error("If X is a list and xlist = TRUE, R also needs to be a list");
@@ -412,7 +423,7 @@ SEXP setcopyv(SEXP x, SEXP val, SEXP rep, SEXP Rinvert, SEXP Rset, SEXP Rind1) {
       const SEXP r = VECTOR_ELT(rep, 0);
       setcopyvLOOPLVEC1
     } else {
-      const SEXP *restrict pr = SEXPPTR(rep);
+      const SEXP *restrict pr = SEXPPTR_RO(rep);
       setcopyvLOOPLVEC
     }
     break;
@@ -455,13 +466,21 @@ SEXP setop_core(SEXP x, SEXP val, SEXP op, SEXP roww) {
 
 #define OPSWITCH(e)                                  \
   switch(o) {                                        \
-  case 1: for(int i = 0; i != n; ++i) px[i] += e;    \
+  case 1:                                            \
+    _Pragma("omp simd")                              \
+    for(int i = 0; i < n; ++i) px[i] += e;           \
     break;                                           \
-  case 2: for(int i = 0; i != n; ++i) px[i] -= e;    \
+  case 2:                                            \
+    _Pragma("omp simd")                              \
+    for(int i = 0; i < n; ++i) px[i] -= e;           \
     break;                                           \
-  case 3: for(int i = 0; i != n; ++i) px[i] *= e;    \
+  case 3:                                            \
+    _Pragma("omp simd")                              \
+    for(int i = 0; i < n; ++i) px[i] *= e;           \
     break;                                           \
-  case 4: for(int i = 0; i != n; ++i) px[i] /= e;    \
+  case 4:                                            \
+    _Pragma("omp simd")                              \
+    for(int i = 0; i < n; ++i) px[i] /= e;           \
     break;                                           \
   default: error("unsupported operation");           \
   }
@@ -512,29 +531,33 @@ if(nv == 1 || nv == n) {
   if((rwl == 0 && nr != nv) || (rwl && nc != nv))
     error("length of vector must match matrix rows/columns or the size of the matrix itself");
 
-#define OPSWITCHMAT(e)                               \
-  switch(o) {                                        \
-  case 1: for(int j = 0, cj; j != nc; ++j)  {        \
-    cj = j * nr;                                     \
-    for(int i = 0; i != nr; ++i) px[cj + i] += e;    \
-  }                                                  \
-  break;                                             \
-  case 2: for(int j = 0, cj; j != nc; ++j)  {        \
-    cj = j * nr;                                     \
-    for(int i = 0; i != nr; ++i) px[cj + i] -= e;    \
-  }                                                  \
-  break;                                             \
-  case 3: for(int j = 0, cj; j != nc; ++j)  {        \
-    cj = j * nr;                                     \
-    for(int i = 0; i != nr; ++i) px[cj + i] *= e;    \
-  }                                                  \
-  break;                                             \
-  case 4: for(int j = 0, cj; j != nc; ++j)  {        \
-    cj = j * nr;                                     \
-    for(int i = 0; i != nr; ++i) px[cj + i] /= e;    \
-  }                                                  \
-  break;                                             \
-  default: error("unsupported operation");           \
+#define OPSWITCHMAT(e)                                 \
+  switch(o) {                                          \
+  case 1: for(int j = 0, cj; j != nc; ++j)  {          \
+    cj = j * nr;                                       \
+    _Pragma("omp simd")                                \
+      for(int i = 0; i < nr; ++i) px[cj + i] += e;     \
+  }                                                    \
+  break;                                               \
+  case 2: for(int j = 0, cj; j != nc; ++j)  {          \
+    cj = j * nr;                                       \
+    _Pragma("omp simd")                                \
+      for(int i = 0; i < nr; ++i) px[cj + i] -= e;     \
+  }                                                    \
+  break;                                               \
+  case 3: for(int j = 0, cj; j != nc; ++j)  {          \
+    cj = j * nr;                                       \
+    _Pragma("omp simd")                                \
+      for(int i = 0; i < nr; ++i) px[cj + i] *= e;     \
+  }                                                    \
+  break;                                               \
+  case 4: for(int j = 0, cj; j != nc; ++j)  {          \
+    cj = j * nr;                                       \
+    _Pragma("omp simd")                                \
+      for(int i = 0; i < nr; ++i) px[cj + i] /= e;     \
+  }                                                    \
+  break;                                               \
+  default: error("unsupported operation");             \
   }
 
 switch(tx) {
@@ -590,10 +613,10 @@ return(x);
 SEXP setop(SEXP x, SEXP val, SEXP op, SEXP roww) {
   // IF x is a list, call function repeatedly..
   if(TYPEOF(x) == VECSXP) {
-    SEXP *px = SEXPPTR(x);
+    const SEXP *px = SEXPPTR_RO(x);
     int lx = length(x);
     if(TYPEOF(val) == VECSXP) { // val is list: must match length(x)
-      SEXP *pv = SEXPPTR(val);
+      const SEXP *pv = SEXPPTR_RO(val);
       if(lx != length(val)) error("length(X) must match length(V)");
       for(int i = 0; i != lx; ++i) setop_core(px[i], pv[i], op, roww);
     } else if (length(val) == 1 || asLogical(roww) == 0) { // val is a scalar or vector but rowwise = FALSE
@@ -624,10 +647,221 @@ SEXP setop(SEXP x, SEXP val, SEXP op, SEXP roww) {
   return setop_core(x, val, op, roww);
 }
 
+
+SEXP replace_outliers(SEXP x, SEXP limits, SEXP value, SEXP single_limit, SEXP set)  {
+  const int ll = length(limits), sl = asInteger(single_limit), l = length(x), setl = asLogical(set);
+  int nprotect = setl == 0;
+  if(ll != 1 && ll != 2) error("'limits' must be length 1 or 2. You supplied limits length %d", ll);
+
+  int clip = 0;
+  if(TYPEOF(value) == STRSXP && strcmp(CHAR(STRING_ELT(value, 0)), "clip") == 0) {
+    value = limits;
+    clip = 1;
+  }
+
+  SEXP res = setl ? x : PROTECT(allocVector(TYPEOF(x), l));
+
+  switch(TYPEOF(x)) {
+    case INTSXP: {
+      if(TYPEOF(limits) != INTSXP) {
+        PROTECT(limits = coerceVector(limits, INTSXP)); ++nprotect;
+      }
+      int *px = INTEGER(x), *pres = INTEGER(res), val = asInteger(value);
+      if(ll == 1) {
+        if(sl == 2 || sl == 3) {
+          int l1 = INTEGER(limits)[0];
+          if(sl == 2) { // minimum
+            #pragma omp simd
+            for (int i = 0; i < l; ++i) pres[i] = px[i] < l1 && px[i] != NA_INTEGER ? val : px[i];
+          } else { // maximum
+            #pragma omp simd
+            for (int i = 0; i < l; ++i) pres[i] = px[i] > l1 ? val : px[i];
+          }
+        }
+      } else { // two-sided
+        int l1 = INTEGER(limits)[0], l2 = INTEGER(limits)[1];
+        if(clip) {
+          #pragma omp simd
+          for (int i = 0; i < l; ++i) pres[i] = px[i] > l2 ? l2 : px[i] < l1 && px[i] != NA_INTEGER ? l1 : px[i];
+        } else {
+          #pragma omp simd
+          for (int i = 0; i < l; ++i) pres[i] = px[i] > l2 || (px[i] < l1 && px[i] != NA_INTEGER) ? val : px[i];
+        }
+      }
+      break;
+    }
+    case REALSXP: {
+      if(TYPEOF(limits) != REALSXP) {
+        PROTECT(limits = coerceVector(limits, REALSXP)); ++nprotect;
+      }
+      double *px = REAL(x), *pres = REAL(res), val = asReal(value);
+      if(ll == 1) {
+        if(sl == 2 || sl == 3) {
+          double l1 = REAL(limits)[0];
+          if(sl == 2) { // minimum
+            #pragma omp simd
+            for (int i = 0; i < l; ++i) pres[i] = px[i] < l1 ? val : px[i];
+          } else { // maximum
+            #pragma omp simd
+            for (int i = 0; i < l; ++i) pres[i] = px[i] > l1 ? val : px[i];
+          }
+        }
+      } else { // two-sided
+        double l1 = REAL(limits)[0], l2 = REAL(limits)[1];
+        if(clip) {
+          #pragma omp simd
+          for (int i = 0; i < l; ++i) pres[i] = px[i] > l2 ? l2 : px[i] < l1 ? l1 : px[i];
+        } else {
+          #pragma omp simd
+          for (int i = 0; i < l; ++i) pres[i] = px[i] > l2 || px[i] < l1 ? val : px[i];
+        }
+      }
+      break;
+    }
+    default: error("Unsupported type '%s'", type2char(TYPEOF(x)));
+  }
+  if(setl == 0) SHALLOW_DUPLICATE_ATTRIB(res, x);
+  UNPROTECT(nprotect);
+  return res;
+}
+
+SEXP na_locf(SEXP x, SEXP Rset) {
+  int n = length(x), copy = asLogical(Rset) == 0;
+  if(isMatrix(x)) warning("na_locf() does not (yet) have explicit support for matrices, i.e., it treats a matrix as a single vector. Use dapply(M, na_locf) if column-wise processing is desired");
+  if(copy) x = PROTECT(shallow_duplicate(x));
+
+  switch (TYPEOF(x)) {
+  case INTSXP:
+  case LGLSXP:
+  {
+    int *data = INTEGER(x);
+    int last = data[0];
+    for (int i = 0; i < n; i++) {
+      if (data[i] == NA_INTEGER) {
+        data[i] = last;
+      } else {
+        last = data[i];
+      }
+    }
+    break;
+  }
+  case REALSXP:
+  {
+    double *data = REAL(x);
+    double last = data[0];
+    for (int i = 0; i < n; i++) {
+      if (ISNAN(data[i])) {
+        data[i] = last;
+      } else {
+        last = data[i];
+      }
+    }
+    break;
+  }
+  case STRSXP:
+  {
+    SEXP *data = STRING_PTR(x);
+    SEXP last = data[0];
+    for (int i = 0; i < n; i++) {
+      if (data[i] == NA_STRING) {
+        data[i] = last;
+      } else {
+        last = data[i];
+      }
+    }
+    break;
+  }
+  case VECSXP:
+  {
+    const SEXP *data = SEXPPTR_RO(x);
+    SEXP last = data[0];
+    for (int i = 0; i < n; i++) {
+      if (length(data[i]) == 0) {
+        SET_VECTOR_ELT(x, i, last);
+      } else {
+        last = data[i];
+      }
+    }
+    break;
+  }
+  default:
+    error("na_locf() does not support type '%s'", type2char(TYPEOF(x)));
+  }
+  UNPROTECT(copy);
+  return x;
+}
+
+SEXP na_focb(SEXP x, SEXP Rset) {
+  int n = length(x), copy = asLogical(Rset) == 0;
+  if(isMatrix(x)) warning("na_focb() does not (yet) have explicit support for matrices, i.e., it treats a matrix as a single vector. Use dapply(M, na_focb) if column-wise processing is desired");
+  if(copy) x = PROTECT(shallow_duplicate(x));
+
+  switch (TYPEOF(x)) {
+  case INTSXP:
+  case LGLSXP:
+  {
+    int *data = INTEGER(x);
+    int last = data[0];
+    for (int i = n; i--; ) {
+      if (data[i] == NA_INTEGER) {
+        data[i] = last;
+      } else {
+        last = data[i];
+      }
+    }
+    break;
+  }
+  case REALSXP:
+  {
+    double *data = REAL(x);
+    double last = data[0];
+    for (int i = n; i--; ) {
+      if (ISNAN(data[i])) {
+        data[i] = last;
+      } else {
+        last = data[i];
+      }
+    }
+    break;
+  }
+  case STRSXP:
+  {
+    SEXP *data = STRING_PTR(x);
+    SEXP last = data[0];
+    for (int i = n; i--; ) {
+      if (data[i] == NA_STRING) {
+        data[i] = last;
+      } else {
+        last = data[i];
+      }
+    }
+    break;
+  }
+  case VECSXP:
+  {
+    const SEXP *data = SEXPPTR_RO(x);
+    SEXP last = data[0];
+    for (int i = n; i--; ) {
+      if (length(data[i]) == 0) {
+        SET_VECTOR_ELT(x, i, last);
+      } else {
+        last = data[i];
+      }
+    }
+    break;
+  }
+  default:
+    error("na_focb() does not support type '%s'", type2char(TYPEOF(x)));
+  }
+  UNPROTECT(copy);
+  return x;
+}
+
+
 SEXP vtypes(SEXP x, SEXP isnum) {
   int tx = TYPEOF(x);
   if(tx != VECSXP) return ScalarInteger(tx);
-  SEXP *px = SEXPPTR(x); // This is ok, even if x contains ALTREP objects..
+  const SEXP *px = SEXPPTR_RO(x); // This is ok, even if x contains ALTREP objects..
   int n = length(x);
   SEXP ans = PROTECT(allocVector(INTSXP, n));
   int *pans = INTEGER(ans);
@@ -750,15 +984,15 @@ SEXP vtypes(SEXP x, SEXP isnum) {
 }
 
 SEXP vlengths(SEXP x, SEXP usenam) {
-  if(TYPEOF(x) != VECSXP) return ScalarInteger(length(x));
+  // if(TYPEOF(x) != VECSXP && TYPEOF(x) != STRSXP) return ScalarInteger(length(x));
   int n = length(x);
   SEXP ans = PROTECT(allocVector(INTSXP, n));
   int *pans = INTEGER(ans);
-  if(ALTREP(x)) {
-    for(int i = 0; i != n; ++i) pans[i] = length(VECTOR_ELT(x, i));
-  } else {
-    SEXP *px = SEXPPTR(x);
+  if(TYPEOF(x) == VECSXP || TYPEOF(x) == STRSXP) {
+    const SEXP *px = SEXPPTR_RO(x);
     for(int i = 0; i != n; ++i) pans[i] = length(px[i]);
+  } else {
+    for(int i = 0; i != n; ++i) pans[i] = 1;
   }
   if(asLogical(usenam)) {
     SEXP nam = getAttrib(x, R_NamesSymbol);
@@ -770,8 +1004,8 @@ SEXP vlengths(SEXP x, SEXP usenam) {
 
 
 // faster version of base::range, which calls both min() and max()
-SEXP frange(SEXP x, SEXP Rnarm) {
-  int l = length(x), narm = asLogical(Rnarm), tx = TYPEOF(x);
+SEXP frange(SEXP x, SEXP Rnarm, SEXP Rfinite) {
+  int l = length(x), narm = asLogical(Rnarm), finite = asLogical(Rfinite), tx = TYPEOF(x);
 
   SEXP out = PROTECT(allocVector(tx, 2));
 
@@ -818,14 +1052,25 @@ SEXP frange(SEXP x, SEXP Rnarm) {
         break;
       }
       double min, max, tmp, *px = REAL(x);
-      if(narm) {
+      if(narm || finite) {
         int j = l-1;
-        while(ISNAN(px[j]) && j!=0) --j;
+        if(finite) while(!R_FINITE(px[j]) && j!=0) --j;
+        else while(ISNAN(px[j]) && j!=0) --j;
         min = max = px[j];
-        if(j != 0) for(int i = j; i--; ) {
-          tmp = px[i];
-          if(min > tmp) min = tmp;
-          if(max < tmp) max = tmp;
+        if(j != 0) {
+          if(finite) {
+            for(int i = j; i--; ) {
+              tmp = px[i];
+              if(min > tmp && tmp > R_NegInf) min = tmp;
+              if(max < tmp && tmp < R_PosInf) max = tmp;
+            }
+          } else {
+            for(int i = j; i--; ) {
+              tmp = px[i];
+              if(min > tmp) min = tmp;
+              if(max < tmp) max = tmp;
+            }
+          }
         }
       } else {
         min = max = px[0];
@@ -897,24 +1142,26 @@ SEXP fdist(SEXP x, SEXP vec, SEXP Rret, SEXP Rnthreads) {
       for(int k = 1; k < nrow; ++k) { // Row vectors to compute distances with
         int nmk = nrow - k;
         double *presk = pres + l - nmk*(nmk+1)/2, // https://en.wikipedia.org/wiki/1_%2B_2_%2B_3_%2B_4_%2B_%E2%8B%AF
-               *pxj = px + k, v, tmp;
+               *pxj = px + k, v;
         for(int j = 0; j != ncol; ++j) { // Elements of the row vector at hand
           v = pxj[-1];
-          for(int i = 0; i != nmk; ++i) { // All remaining rows to compute the distance to
-            tmp = pxj[i] - v;
+          #pragma omp simd
+          for(int i = 0; i < nmk; ++i) { // All remaining rows to compute the distance to
+            double tmp = pxj[i] - v;
             presk[i] += tmp * tmp;
           }
           pxj += nrow;
         }
       }
     } else {
-      double *presk = pres, *pxj, v, tmp;
+      double *presk = pres, *pxj, v;
       for(int k = 1, nmk = nrow; k != nrow; ++k) { // Row vectors to compute distances with
         pxj = px + k; --nmk;
         for(int j = 0; j != ncol; ++j) { // Elements of the row vector at hand
           v = pxj[-1];
-          for(int i = 0; i != nmk; ++i) { // All remaining rows to compute the distance to
-            tmp = pxj[i] - v;
+          #pragma omp simd
+          for(int i = 0; i < nmk; ++i) { // All remaining rows to compute the distance to
+            double tmp = pxj[i] - v;
             presk[i] += tmp * tmp;
           }
           pxj += nrow;
@@ -933,7 +1180,7 @@ SEXP fdist(SEXP x, SEXP vec, SEXP Rret, SEXP Rnthreads) {
         if(nthreads > nrow) nthreads = nrow;
         for (int j = 0; j < ncol; ++j) {
           double *pxj = px + j * nrow, v = pv[j];
-          #pragma omp parallel for num_threads(nthreads)
+          #pragma omp parallel for simd num_threads(nthreads)
           for (int i = 0; i < nrow; ++i) {
             double tmp = pxj[i] - v;
             pres[i] += tmp * tmp;
@@ -941,9 +1188,10 @@ SEXP fdist(SEXP x, SEXP vec, SEXP Rret, SEXP Rnthreads) {
         }
       } else {
         for (int j = 0; j != ncol; ++j) {
-          double *pxj = px + j * nrow, v = pv[j], tmp;
-          for (int i = 0; i != nrow; ++i) {
-            tmp = pxj[i] - v;
+          double *pxj = px + j * nrow, v = pv[j];
+          #pragma omp simd
+          for (int i = 0; i < nrow; ++i) {
+            double tmp = pxj[i] - v;
             pres[i] += tmp * tmp;
           }
         }
@@ -958,9 +1206,9 @@ SEXP fdist(SEXP x, SEXP vec, SEXP Rret, SEXP Rnthreads) {
           dres += tmp * tmp;
         }
       } else {
-        double tmp;
-        for (int i = 0; i != ncol; ++i) {
-          tmp = px[i] - pv[i];
+        #pragma omp simd reduction(+:dres)
+        for (int i = 0; i < ncol; ++i) {
+          double tmp = px[i] - pv[i];
           dres += tmp * tmp;
         }
       }
@@ -972,10 +1220,11 @@ SEXP fdist(SEXP x, SEXP vec, SEXP Rret, SEXP Rnthreads) {
   // Square Root
   if(ret == 1) {
     if(nthreads > 1) {
-      #pragma omp parallel for num_threads(nthreads)
+      #pragma omp parallel for simd num_threads(nthreads)
       for (size_t i = 0; i < l; ++i) pres[i] = sqrt(pres[i]);
     } else {
-      for (size_t i = 0; i != l; ++i) pres[i] = sqrt(pres[i]);
+      #pragma omp simd
+      for (size_t i = 0; i < l; ++i) pres[i] = sqrt(pres[i]);
     }
   }
 
